"""Constants for footing."""

#: The environment variable set when running any footing command. It is set to
#: the name of the command
FOOTING_ENV_VAR = "_FOOTING"

#: The footing config file in each repo
FOOTING_CONFIG_FILE = "footing.yaml"

#: The Github API token environment variable
GITHUB_API_TOKEN_ENV_VAR = "GITHUB_API_TOKEN"

#: The Gitlab API token environment variable
GITLAB_API_TOKEN_ENV_VAR = "GITLAB_API_TOKEN"

#: Footing docs URL
FOOTING_DOCS_URL = "https://github.com/Opus10/footing"

#: The temporary branches used for updates
UPDATE_BRANCH_NAME = "_footing_update"
TEMP_UPDATE_BRANCH_NAME = UPDATE_BRANCH_NAME + "_temp"



"""Creates and initializes a project from a template."""

from __future__ import annotations

import subprocess
import unittest.mock

import cookiecutter.generate as cc_generate
import cookiecutter.hooks as cc_hooks

import footing.check
import footing.constants
import footing.utils


def _patched_run_hook(hook_name, project_dir, context):
    """Used to patch cookiecutter's ``run_hook`` function.

    This patched version ensures that the footing.yaml file is created before
    any cookiecutter hooks are executed
    """
    if hook_name == "post_gen_project":
        with footing.utils.cd(project_dir):
            footing.utils.write_footing_config(
                context["cookiecutter"],
                context["template"],
                context["version"],
            )
    return cc_hooks.run_hook(hook_name, project_dir, context)


def _generate_files(repo_dir, config, template, version):
    """Uses cookiecutter to generate files for the project.

    Monkeypatches cookiecutter's "run_hook" to ensure that the footing.yaml file is
    generated before any hooks run. This is important to ensure that hooks can also
    perform any actions involving footing.yaml
    """
    with unittest.mock.patch("cookiecutter.generate.run_hook", side_effect=_patched_run_hook):
        cc_generate.generate_files(
            repo_dir=repo_dir,
            context={
                "cookiecutter": config,
                "template": template,
                "version": version,
            },
            overwrite_if_exists=False,
            output_dir=".",
        )


@footing.utils.set_cmd_env_var("setup")
def setup(template: str, version: str | None = None) -> None:
    """Sets up a new project from a template

    Note that the `footing.constants.FOOTING_ENV_VAR` is set to 'setup' during the duration
    of this function.

    Args:
        template: The git path to a template
        version: The version of the template to use when updating. Defaults
            to the latest version
    """
    footing.check.not_in_git_repo()

    repo_path = footing.utils.get_repo_path(template)
    msg = (
        "You will be prompted for the parameters of your new project."
        " Please read the docs at https://github.com/{} before entering parameters."
    ).format(repo_path)
    print(msg)

    cc_repo_dir, config = footing.utils.get_cookiecutter_config(template, version=version)

    if not version:
        with footing.utils.cd(cc_repo_dir):
            ret = footing.utils.shell("git rev-parse HEAD", stdout=subprocess.PIPE)
            version = ret.stdout.decode("utf-8").strip()

    _generate_files(repo_dir=cc_repo_dir, config=config, template=template, version=version)



"""Updates a footing project with the latest template."""

from __future__ import annotations

import json
import os
import shutil
import subprocess
import tempfile
import textwrap

import cookiecutter.main as cc_main
import cookiecutter.vcs as cc_vcs

import footing.check
import footing.constants
import footing.forge
import footing.utils


def _cookiecutter_configs_have_changed(template, old_version, new_version):
    """Given an old version and new version, check if the cookiecutter.json files have changed

    When the cookiecutter.json files change, it means the user will need to be prompted for
    new context

    Args:
        template (str): The git path to the template
        old_version (str): The git SHA of the old version
        new_version (str): The git SHA of the new version

    Returns:
        bool: True if the cookiecutter.json files have been changed in the old and new versions
    """
    with tempfile.TemporaryDirectory() as clone_dir:
        repo_dir = cc_vcs.clone(template, old_version, clone_dir)
        old_config = json.load(open(os.path.join(repo_dir, "cookiecutter.json")))
        subprocess.check_call(
            "git checkout %s" % new_version,
            cwd=repo_dir,
            shell=True,
            stderr=subprocess.PIPE,
        )
        new_config = json.load(open(os.path.join(repo_dir, "cookiecutter.json")))

    return old_config != new_config


def _apply_template(template, target, *, checkout, extra_context):
    """Apply a template to a temporary directory and then copy results to target."""
    with tempfile.TemporaryDirectory() as tempdir:
        repo_dir = cc_main.cookiecutter(
            template,
            checkout=checkout,
            no_input=True,
            output_dir=tempdir,
            extra_context=extra_context,
        )
        for item in os.listdir(repo_dir):
            src = os.path.join(repo_dir, item)
            dst = os.path.join(target, item)
            if os.path.isdir(src):
                if os.path.exists(dst):
                    shutil.rmtree(dst)
                shutil.copytree(src, dst)
            else:
                if os.path.exists(dst):
                    os.remove(dst)
                shutil.copy2(src, dst)


def _get_latest_template_version(template):
    """Obtains the latest template version from the appropriate git forge"""
    client = footing.forge.from_path(template)
    return client.get_latest_template_version(template)


@footing.utils.set_cmd_env_var("update")
def up_to_date(version: str | None = None) -> bool:
    """Checks if a footing project is up to date with the repo

    Note that the `footing.constants.FOOTING_ENV_VAR` is set to 'update' for the duration of this
    function.

    Args:
        version: Update against this git SHA or branch of the template

    Returns:
        True if up to date with ``version`` (or latest version), False otherwise

    Raises:
        `NotInGitRepoError`: When running outside of a git repo
        `InvalidFootingProjectError`: When not inside a valid footing repository
    """
    footing.check.in_git_repo()
    footing.check.is_footing_project()

    footing_config = footing.utils.read_footing_config()
    old_template_version = footing_config["_version"]
    new_template_version = version or _get_latest_template_version(footing_config["_template"])

    return new_template_version == old_template_version


def _needs_new_cc_config_for_update(old_template, old_version, new_template, new_version):
    """
    Given two templates and their respective versions, return True if a new cookiecutter
    config needs to be obtained from the user
    """
    if old_template != new_template:
        return True
    else:
        return _cookiecutter_configs_have_changed(new_template, old_version, new_version)


@footing.utils.set_cmd_env_var("update")
def update(
    old_template: str | None = None,
    old_version: str | None = None,
    new_template: str | None = None,
    new_version: str | None = None,
    enter_parameters: bool = False,
) -> bool:
    """Updates the footing project to the latest template

    Proceeeds in the following steps:

    1. Ensure we are inside the project repository
    2. Obtain the latest version of the package template
    3. If the package is up to date with the latest template, return
    4. If not, create an empty template branch with a new copy of the old template
    5. Create an update branch from HEAD and merge in the new template copy
    6. Create a new copy of the new template and merge into the empty template branch
    7. Merge the updated empty template branch into the update branch
    8. Ensure footing.yaml reflects what is in the template branch
    9. Remove the empty template branch

    Note that the `footing.constants.FOOTING_ENV_VAR` is set to 'update' for the
    duration of this function.

    Two branches will be created during the update process, one named
    ``_footing_update`` and one named ``_footing_update_temp``. At the end of
    the process, ``_footing_update_temp`` will be removed automatically. The
    work will be left in ``_footing_update`` in an uncommitted state for
    review. The update will fail early if either of these branches exist
    before the process starts.

    Args:
        old_template: The old template from which to update. Defaults to the template in
            footing.yaml.
        old_version: The old version of the template. Defaults to the version in footing.yaml.
        new_template: The new template for updating. Defaults to the template in footing.yaml
        new_version: The new version of the new template to update. Defaults to the latest version
            of the new template.
        enter_parameters: Force entering template parameters for the project

    Raises:
        `NotInGitRepoError`: When not inside of a git repository
        `InvalidFootingProjectError`: When not inside a valid footing repository
        `InDirtyRepoError`: When an update is triggered while the repo is in a dirty state
        `ExistingBranchError`: When an update is triggered and there is an existing
            update branch

    Returns:
        True if update was performed or False if template was already up to date
    """
    update_branch = footing.constants.UPDATE_BRANCH_NAME
    temp_update_branch = footing.constants.TEMP_UPDATE_BRANCH_NAME

    footing.check.in_git_repo()
    footing.check.in_clean_repo()
    footing.check.is_footing_project()
    footing.check.not_has_branch(update_branch)
    footing.check.not_has_branch(temp_update_branch)

    footing_config = footing.utils.read_footing_config()
    old_template = old_template or footing_config["_template"]
    new_template = new_template or footing_config["_template"]
    old_version = old_version or footing_config["_version"]
    new_version = new_version or _get_latest_template_version(new_template)

    if new_template == old_template and new_version == old_version and not enter_parameters:
        print("No updates have happened to the template, so no files were updated")
        return False

    print("Creating branch {} for processing the update".format(update_branch))
    footing.utils.shell("git checkout -b {}".format(update_branch), stderr=subprocess.DEVNULL)

    print("Creating temporary working branch {}".format(temp_update_branch))
    footing.utils.shell(
        "git checkout --orphan {}".format(temp_update_branch),
        stderr=subprocess.DEVNULL,
    )
    footing.utils.shell("git rm -rf .", stdout=subprocess.DEVNULL)
    _apply_template(old_template, ".", checkout=old_version, extra_context=footing_config)
    footing.utils.shell("git add .")
    footing.utils.shell(
        'git commit --no-verify -m "Initialize template from version {}"'.format(old_version),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    print("Merge old template history into update branch.")
    footing.utils.shell("git checkout {}".format(update_branch), stderr=subprocess.DEVNULL)
    footing.utils.shell(
        "git merge -s ours --no-edit --allow-unrelated-histories {}".format(temp_update_branch),
        stderr=subprocess.DEVNULL,
    )

    print("Update template in temporary branch.")
    footing.utils.shell("git checkout {}".format(temp_update_branch), stderr=subprocess.DEVNULL)
    footing.utils.shell("git rm -rf .", stdout=subprocess.DEVNULL)

    # If the cookiecutter.json files have changed or the templates have changed,
    # the user will need to re-enter the cookiecutter config
    needs_new_cc_config = _needs_new_cc_config_for_update(
        old_template, old_version, new_template, new_version
    )
    if needs_new_cc_config:
        if old_template != new_template:
            cc_config_input_msg = (
                "You will be prompted for the parameters of the new template."
                " Please read the docs at https://github.com/{} before entering parameters."
                " Press enter to continue"
            ).format(footing.utils.get_repo_path(new_template))
        else:
            cc_config_input_msg = (
                "A new template variable has been defined in the updated template."
                " You will be prompted to enter all of the variables again. Variables"
                " already configured in your project will have their values set as"
                " defaults. Press enter to continue"
            )

        input(cc_config_input_msg)

    # Even if there is no detected need to re-enter the cookiecutter config, the user
    # can still re-enter config parameters with the "enter_parameters" flag
    if needs_new_cc_config or enter_parameters:
        _, footing_config = footing.utils.get_cookiecutter_config(
            new_template, default_config=footing_config, version=new_version
        )

    _apply_template(new_template, ".", checkout=new_version, extra_context=footing_config)
    footing.utils.write_footing_config(footing_config, new_template, new_version)

    footing.utils.shell("git add .")
    footing.utils.shell(
        'git commit --no-verify -m "Update template to version {}"'.format(new_version),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    print("Merge updated template into update branch.")
    footing.utils.shell("git checkout {}".format(update_branch), stderr=subprocess.DEVNULL)
    footing.utils.shell(
        "git merge --no-commit {}".format(temp_update_branch),
        check=False,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    # The footing.yaml file should always reflect what is in the new template
    footing.utils.shell(
        "git checkout --theirs {}".format(footing.constants.FOOTING_CONFIG_FILE),
        check=False,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    print("Remove temporary template branch {}".format(temp_update_branch))
    footing.utils.shell(
        "git branch -D {}".format(temp_update_branch),
        stdout=subprocess.DEVNULL,
    )

    print(
        textwrap.dedent(
            """\
        Updating complete!

        Please review the changes with "git status" for any errors or
        conflicts. Once you are satisfied with the changes, add, commit,
        push, and open a PR with the branch {}
    """
        ).format(update_branch)
    )
    return True



"""
The footing CLI contains commands for setting up, listing, and updating projects.

Commands:

* `footing setup` - Sets up a new project
* `footing ls` - Lists all templates and projects created with those templates
* `footing update` - Updates the project to the latest template version
* `footing clean` - Cleans up any temporary resources used by footing
* `footing switch` - Switch a project to a different template
"""

import click
import pkg_resources

import footing
import footing.clean
import footing.exceptions
import footing.ls
import footing.setup
import footing.update


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("--version", is_flag=True, help="Show version")
def main(ctx, version):
    if version:
        print("footing {}".format(pkg_resources.get_distribution("footing").version))
    elif not ctx.invoked_subcommand:
        print(ctx.get_help())


@main.command()
@click.argument("template", nargs=1, required=True)
@click.option(
    "-v",
    "--version",
    default=None,
    help="Git SHA or branch of template to use for creation",
)
def setup(template, version):
    """
    Setup new project. Takes a git path to the template as returned
    by "footing ls". In order to start a project from a
    particular version (instead of the latest), use the "-v" option.
    """
    footing.setup.setup(template, version=version)


@main.command()
@click.option("-c", "--check", is_flag=True, help="Check to see if up to date")
@click.option(
    "-e",
    "--enter-parameters",
    is_flag=True,
    help="Enter template parameters on update",
)
@click.option(
    "-v",
    "--version",
    default=None,
    help="Git SHA or branch of template to use for update",
)
def update(check, enter_parameters, version):
    """
    Update package with latest template. Must be inside of the project
    folder to run.

    Using "-e" will prompt for re-entering the template parameters again
    even if the project is up to date.

    Use "-v" to update to a particular version of a template.

    Using "-c" will perform a check that the project is up to date
    with the latest version of the template (or the version specified by "-v").
    No updating will happen when using this option.
    """
    if check:
        if footing.update.up_to_date(version=version):
            print("Footing package is up to date")
        else:
            msg = (
                "This footing package is out of date with the latest template."
                ' Update your package by running "footing update" and commiting changes.'
            )
            raise footing.exceptions.NotUpToDateWithTemplateError(msg)
    else:
        footing.update.update(new_version=version, enter_parameters=enter_parameters)


@main.command()
@click.argument("forge", nargs=1, required=True)
@click.argument("template", nargs=1, required=False)
@click.option(
    "-l",
    "--long-format",
    is_flag=True,
    help="Print extended information about results",
)
def ls(forge, template, long_format):
    """
    List packages created with footing. Enter a git forge path, such as a Github
    user or Gitlab group URL, to list all templates under the forge.
    Provide the template path as the second argument to list all projects
    that have been started with that template.

    Use "-l" to print the repository descriptions of templates
    or projects.
    """
    results = footing.ls.ls(forge, template=template)
    for ssh_path, description in results.items():
        if long_format:
            print(ssh_path, "-", description)
        else:
            print(ssh_path)


@main.command()
def clean():
    """
    Cleans temporary resources created by footing, such as the footing update branch
    """
    footing.clean.clean()


@main.command()
@click.argument("template", nargs=1, required=True)
@click.option(
    "-v",
    "--version",
    default=None,
    help="Git SHA or branch of template to use for update",
)
def switch(template, version):
    """
    Switch a project's template to a different template.
    """
    footing.update.update(new_template=template, new_version=version)



"""Footing exceptions."""


class Error(Exception):
    """The top-level error for footing"""


class InGitRepoError(Error):
    """Thrown when running inside of a git repository"""


class NotInGitRepoError(Error):
    """Thrown when not running inside of a git repo"""


class InDirtyRepoError(Error):
    """Thrown when running in a dirty git repo"""


class InvalidFootingProjectError(Error):
    """Thrown when the repository was not created with footing"""


class NotUpToDateWithTemplateError(Error):
    """Thrown when a footing project is not up to date with the template"""


class CheckRunError(Error):
    """When running ``footing update --check`` errors"""


class InvalidEnvironmentError(Error):
    """Thrown when required environment variables are not set"""


class InvalidForgeError(Error):
    """An invalid forge was passed to ls."""


class InvalidTemplatePathError(Error):
    """Thrown when a template path is not a Github SSH path"""


class ExistingBranchError(Error):
    """Thrown when a specifically named branch exists or doesn't exist as expected."""


class InvalidCurrentBranchError(Error):
    """Thrown when a command cannot run because of the current git branch"""


class InvalidGitlabGroupError(Error):
    """Thrown when an invalid Gitlab group is provided"""



"""Functions for cleaning up temporary resources used by footing."""

import subprocess

import footing.check
import footing.constants
import footing.exceptions
import footing.utils


def _get_current_branch():
    """Determine the current git branch"""
    result = footing.utils.shell("git rev-parse --abbrev-ref HEAD", stdout=subprocess.PIPE)
    return result.stdout.decode("utf8").strip()


def clean() -> None:
    """Cleans up temporary resources

    Tries to clean up:

    1. The temporary update branch used during ``footing update``
    2. The primary update branch used during ``footing update``
    """
    footing.check.in_git_repo()

    current_branch = _get_current_branch()
    update_branch = footing.constants.UPDATE_BRANCH_NAME
    temp_update_branch = footing.constants.TEMP_UPDATE_BRANCH_NAME

    if current_branch in (update_branch, temp_update_branch):
        err_msg = (
            'You must change from the "{}" branch since it will be deleted during cleanup'
        ).format(current_branch)
        raise footing.exceptions.InvalidCurrentBranchError(err_msg)

    if footing.check._has_branch(update_branch):
        footing.utils.shell("git branch -D {}".format(update_branch))
    if footing.check._has_branch(temp_update_branch):
        footing.utils.shell("git branch -D {}".format(temp_update_branch))



"""
Utilities for accessing and traversing different git forges, along with pulling down
remote templates.

Currently Github and Gitlab are supported.
"""

from __future__ import annotations

import abc
import collections
import os
import re
import subprocess
from urllib.parse import urlparse

import gitlab
import gitlab.const
import requests
import tldextract

import footing.check
import footing.constants
import footing.exceptions
import footing.utils


def from_path(path):
    """
    Given a forge path, such as Github or Gitlab, return a client for accessing
    the repository information.

    Args:
        path (str): A path under which templates are stored.
            For example, a Github organization or user (e.g. github.com/Organization)
            or a Gitlab group (e.g. gitlab.com/my/group).
    """
    if "github.com" in path:
        return Github()
    elif "gitlab.com" in path:
        return Gitlab()
    else:
        raise footing.exceptions.InvalidForgeError(
            "Invalid forge provided. Must provide either a Github user/organization"
            " (e.g. github.com/UserName) or"
            " a Gitlab group (e.g. gitlab.com/my/group).",
        )


def get_name_from_ssh_path(template_path):
    matches = re.search(r"\/([^/]+)\.git$", template_path)
    return matches.group(1) if matches else ""


def _get_latest_template_version_w_ssh(template):
    """
    Tries to obtain the latest template version using an SSH key
    """
    cmd = "git ls-remote {} | grep HEAD | cut -f1".format(template)
    ret = footing.utils.shell(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stderr = ret.stderr.decode("utf-8").strip()
    stdout = ret.stdout.decode("utf-8").strip()
    if stderr and not stdout:
        raise RuntimeError(
            ('An unexpected error happened when running "{}". (stderr="{}"').format(cmd, stderr)
        )
    return stdout


class Forge(metaclass=abc.ABCMeta):
    """The base class for all git forges.

    Forges must implement both ``ls`` for listing templates/projects
    and ``_get_latest_template_version`` for finding the latest version
    of a template. The ``api_token_env_var_name`` property must also
    be configured.
    """

    @abc.abstractmethod
    def ls(self, path, template=None) -> dict[str, str]:
        """Implements ls for the forge"""
        pass

    @property
    @abc.abstractmethod
    def api_token_env_var_name(self) -> str:
        """Returns the environment variable name for configuring an API token"""
        pass

    def get_latest_template_version(self, template):
        """Retrieves the latest template SHA

        Returns:
            str: The latest template version
        """
        try:
            latest_version = _get_latest_template_version_w_ssh(template)
        except (subprocess.CalledProcessError, RuntimeError):
            try:
                latest_version = self._get_latest_template_version(template)
            except (
                requests.exceptions.RequestException,
                footing.exceptions.InvalidEnvironmentError,
            ) as exc:
                raise footing.exceptions.CheckRunError(
                    (
                        'Could not obtain the latest template version of "{}"'
                        " using git SSH key or the configured {} API token."
                        ' Set either a "{}" environment variable'
                        " with access to the template or obtain permission so that"
                        " the git SSH key can access it."
                    ).format(
                        template,
                        self.__class__.__name__,
                        self.api_token_env_var_name,
                    )
                ) from exc

        return latest_version

    @abc.abstractmethod
    def _get_latest_template_version(self, template):
        """Finds the latest version of a template using an API

        By default, the latest version of a template is used with standard
        git calls and SSH auth. However, one must implement this method
        as a fallback in case only API access is available.
        """
        pass


class Github(Forge):
    """A Github forge"""

    @property
    def api_token_env_var_name(self):
        return footing.constants.GITHUB_API_TOKEN_ENV_VAR

    def _call_api(self, verb, url, **request_kwargs):
        """Perform a github API call

        Args:
            verb (str): Can be "post", "put", or "get"
            url (str): The base URL with a leading slash for Github API (v3)
            auth (str or HTTPBasicAuth): A Github API token or a HTTPBasicAuth object
        """
        footing.check.has_env_vars(footing.constants.GITHUB_API_TOKEN_ENV_VAR)
        api_token = os.environ[footing.constants.GITHUB_API_TOKEN_ENV_VAR]
        api = "https://api.github.com{}".format(url)
        auth_headers = {"Authorization": "token {}".format(api_token)}
        headers = {**auth_headers, **request_kwargs.pop("headers", {})}
        return getattr(requests, verb)(api, headers=headers, **request_kwargs)

    def _get(self, url, **request_kwargs):
        """Github API get"""
        return self._call_api("get", url, **request_kwargs)

    def _parse_link_header(self, headers):
        """A utility function that parses Github's link header for pagination."""
        links = {}
        if "link" in headers:
            link_headers = headers["link"].split(", ")
            for link_header in link_headers:
                (url, rel) = link_header.split("; ")
                url = url[1:-1]
                rel = rel[5:-1]
                links[rel] = url
        return links

    def _code_search(self, query: str, forge: str | None = None) -> dict[str, dict[str, str]]:
        """Performs a Github API code search

        Args:
            query: The query sent to Github's code search
            forge: The root being searched in Github

        Returns:
            A dictionary of repository information keyed on the git SSH url

        Raises:
            `InvalidForgeError`: When ``forge`` is invalid
        """
        headers = {"Accept": "application/vnd.github.v3.text-match+json"}

        resp = self._get(
            "/search/code",
            params={"q": query, "per_page": 100},
            headers=headers,
        )

        if resp.status_code == requests.codes.unprocessable_entity and forge:
            raise footing.exceptions.InvalidForgeError('Invalid Github forge - "{}"'.format(forge))
        resp.raise_for_status()

        resp_data = resp.json()

        repositories: dict[str, dict[str, str]] = collections.defaultdict(dict)
        while True:
            repositories.update(
                {
                    "git@github.com:{}.git".format(repo["repository"]["full_name"]): repo[
                        "repository"
                    ]
                    for repo in resp_data["items"]
                }
            )

            next_url = self._parse_link_header(resp.headers).get("next")
            if next_url:
                resp = requests.get(next_url, headers=headers)
                resp.raise_for_status()
                resp_data = resp.json()
            else:
                break

        return repositories

    def _get_latest_template_version(self, template):
        """Tries to obtain the latest template version with the Github API"""
        repo_path = footing.utils.get_repo_path(template)
        api = "/repos/{}/commits".format(repo_path)

        last_commit_resp = self._get(api, params={"per_page": 1})
        last_commit_resp.raise_for_status()

        content = last_commit_resp.json()
        assert len(content) == 1, "Unexpected Github API response"
        return content[0]["sha"]

    def ls(self, path, template=None):
        """Return a list of repositories under the forge path or the template (if provided)."""

        path_parts = path.strip().split("/")
        if path_parts[-1] == "":
            user_or_org = path_parts[-2]
        else:
            user_or_org = path_parts[-1]

        if template:
            footing.check.is_git_ssh_path(template)
            template_repo_name = get_name_from_ssh_path(template)

            search_q = "user:{} filename:{} {}".format(
                user_or_org,
                footing.constants.FOOTING_CONFIG_FILE,
                template_repo_name,
            )
        else:
            search_q = "user:{} filename:cookiecutter.json".format(user_or_org)

        results = self._code_search(search_q, forge=path)
        return collections.OrderedDict(
            sorted(
                [
                    (key, value["description"] or "(no description found)")
                    for key, value in results.items()
                ]
            )
        )


class Gitlab(Forge):
    """A Gitlab forge"""

    @property
    def api_token_env_var_name(self):
        return footing.constants.GITLAB_API_TOKEN_ENV_VAR

    def get_client(self, gitlab_url):
        footing.check.has_env_vars(self.api_token_env_var_name)
        api_token = os.environ[self.api_token_env_var_name]
        return gitlab.Gitlab(url=gitlab_url, private_token=api_token)

    def _get_gitlab_url_and_repo_path(self, template):
        """Given a template, return a gitlab url and a repo path"""
        repo_path = footing.utils.get_repo_path(template)

        # Figure out the top-level gitlab domain. Note that it's assumed
        # the template is an SSH path in the format of git@domain:path
        domain = template.split(":")[0].split("@")[1]
        gitlab_url = "https://" + domain

        return gitlab_url, repo_path

    def _get_latest_template_version(self, template):  # pragma: no cover
        """Tries to obtain the latest template version with the Gitlab API"""
        gitlab_url, repo_path = self._get_gitlab_url_and_repo_path(template)

        gl = self.get_client(gitlab_url)
        project = gl.projects.get(repo_path)
        sha = project.commits.list()[0].id  # type: ignore

        return sha

    def _get_gitlab_url_and_group(self, forge):
        """Given a forge, return a gitlab url and group"""
        if not forge.startswith("http"):
            forge = "https://" + forge

        url_parts = urlparse(forge)
        gitlab_url = url_parts.scheme + "://" + url_parts.netloc
        group = url_parts.path.strip("/")

        # If users are listing templates on gitlab.com and not a self-hosted gitlab,
        # do not allow them to query the root gitlab.com
        is_self_hosted = tldextract.extract(forge).domain != "gitlab"

        if not group and not is_self_hosted:
            raise footing.exceptions.InvalidGitlabGroupError(
                "Must provide a gitlab group, for example gitlab.com/group"
            )

        return gitlab_url, group

    def ls(self, path, template=None):  # pragma: no cover
        """Return a list of repositories under the forge path or the template (if provided)."""
        gitlab_url, group = self._get_gitlab_url_and_group(path)

        gl = self.get_client(gitlab_url)
        if group:
            # Search under a group if one is specified
            gl = gl.groups.get(group)

        # Search for either templates (with cookiecutter.json) or projects that have been made
        # from the template. Note - advanced search must be turned on for the Gitlab instance
        if not template:
            results = gl.search(  # type: ignore
                gitlab.const.SEARCH_SCOPE_BLOBS, search="filename:cookiecutter.json"
            )
        else:
            results = gl.search(  # type: ignore
                gitlab.const.SEARCH_SCOPE_BLOBS, search="{} filename:footing.yaml".format(template)
            )

        # Fetch projects associated with search results
        gl = self.get_client(gitlab_url)
        projects = [gl.projects.get(r["project_id"]) for r in results]  # type: ignore

        results: dict[str, str] = collections.OrderedDict(
            sorted(
                [
                    (
                        p.ssh_url_to_repo,
                        p.description or "(no description found)",
                    )
                    for p in projects
                ]
            )
        )
        return results



"""Shared footing utilities."""

import contextlib
import functools
import os
import subprocess

import cookiecutter.config as cc_config
import cookiecutter.generate as cc_generate
import cookiecutter.prompt as cc_prompt
import cookiecutter.repository as cc_repository
import yaml

import footing.constants
import footing.exceptions


def get_repo_path(template):
    """Given a git SSH path (e.g git@github.com:owner/repo.git), return the repo path

    The repo path is in the form of "owner/repo"
    """
    return template[:-4].split(":")[1]


def shell(cmd, check=True, stdin=None, stdout=None, stderr=None):
    """Runs a subprocess shell with check=True by default"""
    return subprocess.run(cmd, shell=True, check=check, stdin=stdin, stdout=stdout, stderr=stderr)


@contextlib.contextmanager
def cd(path):
    """A context manager for changing into a directory"""
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def read_footing_config():
    """Reads the footing YAML configuration file in the repository"""
    with open(footing.constants.FOOTING_CONFIG_FILE) as footing_config_file:
        return yaml.load(footing_config_file, Loader=yaml.SafeLoader)


def write_footing_config(footing_config, template, version):
    """Writes the footing YAML configuration"""
    with open(footing.constants.FOOTING_CONFIG_FILE, "w") as footing_config_file:
        versioned_config = {
            **footing_config,
            **{"_version": version, "_template": template},
        }
        yaml.dump(versioned_config, footing_config_file, Dumper=yaml.SafeDumper)


def get_cookiecutter_config(template, default_config=None, version=None):
    """Obtains the configuration used for cookiecutter templating

    Args:
        template: Path to the template
        default_config (dict, optional): The default configuration
        version (str, optional): The git SHA or branch to use when
            checking out template. Defaults to latest version

    Returns:
        tuple: The cookiecutter repo directory and the config dict
    """
    default_config = default_config or {}
    config_dict = cc_config.get_user_config()
    repo_dir, _ = cc_repository.determine_repo_dir(
        template=template,
        abbreviations=config_dict["abbreviations"],
        clone_to_dir=config_dict["cookiecutters_dir"],
        checkout=version,
        no_input=True,
    )
    context_file = os.path.join(repo_dir, "cookiecutter.json")
    context = cc_generate.generate_context(
        context_file=context_file,
        default_context={**config_dict["default_context"], **default_config},
    )
    return repo_dir, cc_prompt.prompt_for_config(context)


def set_cmd_env_var(value):
    """Decorator that sets the footing command env var to value"""

    def func_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            previous_cmd_env_var = os.getenv(footing.constants.FOOTING_ENV_VAR)
            os.environ[footing.constants.FOOTING_ENV_VAR] = value
            try:
                ret_val = function(*args, **kwargs)
            finally:
                if previous_cmd_env_var is None:
                    del os.environ[footing.constants.FOOTING_ENV_VAR]
                else:
                    os.environ[footing.constants.FOOTING_ENV_VAR] = previous_cmd_env_var

            return ret_val

        return wrapper

    return func_decorator



"""Lists all footing templates and projects spun up with those templates."""

from __future__ import annotations

import footing.forge
import footing.utils


@footing.utils.set_cmd_env_var("ls")
def ls(forge: str, template: str | None = None) -> dict[str, str]:
    """Lists all templates under a root path or list all projects spun up under
    a root path and a template path.

    The ``root`` path must be either a Github organization/user (e.g. github.com/organization)
    or a Gitlab group (e.g. gitlab.com/my/group).

    Note that the `footing.constants.FOOTING_ENV_VAR` is set to 'ls' for the duration of this
    function.

    Args:
        forge: A root git storage path.  For example, a Github organization
            (github.com/Organization) or a gitlab group (gitlab.com/my/group).
        template: An optional template path. If provided, the
            returned values are projects under ``root`` created using the template.

    Returns:
        xA dictionary of repository information keyed on the url.

    Raises:
        `InvalidForgeError`: When ``forge`` is invalid
    """
    client = footing.forge.from_path(forge)
    return client.ls(forge, template)



"""Utilities for performing checks and throwing useful error messages."""

import os
import subprocess

import footing.constants
import footing.exceptions
import footing.utils


def is_git_ssh_path(template_path):
    """Raises a `InvalidTemplatePathError` if ``template_path`` is not a git SSH url

    Note that the git SSH url must be in the form as provided from Github or from
    ``footing ls``. For example, ``git@github.com:user/template.git``.
    """
    if not template_path.startswith("git@github.com:") or not template_path.endswith(".git"):
        raise footing.exceptions.InvalidTemplatePathError(
            'The template path must be a git SSH url (e.g. "git@github.com:user/template.git")'
        )


def _in_git_repo():
    """Returns True if inside a git repo, False otherwise"""
    ret = footing.utils.shell("git rev-parse", stderr=subprocess.DEVNULL, check=False)
    return ret.returncode == 0


def in_git_repo():
    """Raises `NotInGitRepoError` if not inside a git repository"""
    if not _in_git_repo():
        msg = "Must be inside git repository of project to run this command."
        raise footing.exceptions.NotInGitRepoError(msg)


def not_in_git_repo():
    """Raises `InGitRepoError` if inside of a git repository"""
    if _in_git_repo():
        msg = "Cannot run inside of a git repository. Change to another directory."
        raise footing.exceptions.InGitRepoError(msg)


def _in_clean_repo():
    """Returns True if the git repo is not dirty, False otherwise"""
    ret = footing.utils.shell("git diff-index --quiet HEAD --", check=False)
    return ret.returncode == 0


def in_clean_repo():
    """Raises `InDirtyRepoError` if inside a dirty repository"""
    if not _in_clean_repo():
        msg = "Cannot run inside of a dirty git repository. Stash or commit changes."
        raise footing.exceptions.InDirtyRepoError(msg)


def _has_branch(branch):
    """Return True if the target branch exists."""
    ret = footing.utils.shell(
        "git rev-parse --verify {}".format(branch),
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        check=False,
    )
    return ret.returncode == 0


def not_has_branch(branch):
    """Raises `ExistingBranchError` if the specified branch exists."""
    if _has_branch(branch):
        msg = "Cannot proceed while {} branch exists; remove and try again.".format(branch)
        raise footing.exceptions.ExistingBranchError(msg)


def has_env_vars(*env_vars):
    """Raises `InvalidEnvironmentError` when one isnt set"""
    for env_var in env_vars:
        if not os.environ.get(env_var):
            msg = (
                "Must set {} environment variable. View docs for setting up environment at {}"
            ).format(env_var, footing.constants.FOOTING_DOCS_URL)
            raise footing.exceptions.InvalidEnvironmentError(msg)


def is_footing_project():
    """Raises `InvalidFootingProjectError` if repository is not a footing project"""
    if not os.path.exists(footing.constants.FOOTING_CONFIG_FILE):
        msg = "No {} file found in repository.".format(footing.constants.FOOTING_CONFIG_FILE)
        raise footing.exceptions.InvalidFootingProjectError(msg)
